%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Bousi-Prolog command-line shell

:- module(bplShell, [
		start_bpl_shell/0       %
   ]).

:- use_module(bplHelp).
:- use_module(translator).
:- use_module(evaluator).
:- use_module(flags).
:- use_module(foreign).
:- use_module(utilities).

:- use_module(library(lists)).
:- use_module(library(system)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Command-line shell
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% start_bpl_shell
%
%     Initializes the environment and launches Bousi-Prolog
%     command-line shell.
%

start_bpl_shell :-
	load_history,
	set_system_predicates,
	flags:reset_bpl_flags,
	evaluator:load_tpl(''),
	bpl_shell_loop.


%% bpl_shell_loop
%
%     Command-line shell main loop. This predicate implements a failure-
%     driven loop that reads a command from standard input, tries to
%     execute it and then repeats the same process indefinitely.
%

bpl_shell_loop :-
	repeat,
	bpl_shell,
	!.


%% bpl_shell
%
%     Helper predicate called by bpl_shell_loop/0 which reads a single
%     command for input, executes it and then fails (note that
%     bpl_shell_loop/0 implements a failure-driven loop).
%
%     @see bpl_shell_loop/0
%

bpl_shell :-
	bpl_prompt(Prompt),
	read_shell_line(Prompt, String, Arguments),
	catch((
		% Converts the read line into an executable term
		translate_command(String, Arguments, Command),
		% Executes the command or query
		Command,
		(Command \== true -> nl ; true)
	% (catcher)
	), Error, (
		(Error = translate_error(ErrorMessage) ->
			% Invalid command or query
			write(ErrorMessage), nl, nl
		;
			% Exception thrown by SWI-Prolog
			print_message(error, Error), nl
		)
	)),
	% This forced fail makes Prolog to go back to the repeat/0
	% predicate of bpl_shell_loop/0, thus repeating this predicate
	% again without nesting calls
	!,
	fail.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for reading and translating Bousi-Prolog commands
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% read_shell_line(+Prompt, ?String, ?Arguments)
%
%     Displays Prompt to user, reads a line from standard input and
%     returns the text of the full line in String and a list with the
%     substrings of the line delimited by whitespace in Arguments.
%     
%

read_shell_line(Prompt, String, Arguments) :-
	foreign:ext_read_shell_line(Prompt, String, ArgumentsWithQuotes),
	utilities:remove_quotes(ArgumentsWithQuotes, Arguments).


%% translate_command(+String, +Arguments, -Command)
%
%     Takes a line read from Bousi-Prolog shell and returns it as an
%     executable term in Command. String must contain the full line
%     read from shell, whereas Arguments must be a list with the
%     substrings of the line delimited by whitespace.
%
%     @throws translate_error(Message) String doesn't contain a Bousi-
%      Prolog shell command nor a query.
%

translate_command(_String, [], true).
	% String contains an empty command

translate_command(_String, Arguments, Command) :-
	% Checks if string contains a Bousi-Prolog shell command
	phrase(command(Command), Arguments).

translate_command(String, Arguments, Command) :-
	% Checks if string contains the special solve command
	Arguments = [sv|_],
	% Extracts query from string and parses it
	sub_atom(String, CommandStart, _, _, 'sv'),
	QueryStart is CommandStart + 2,
	sub_atom(String, QueryStart, _, 0, QueryString),
	translator:translate_query(QueryString, Query, Bindings, Degree),
	Command = sv(Query, Bindings, Degree).

translate_command(_String, Arguments, _Command) :-
	% Checks if string contains an invalid Bousi-Prolog shell command
	Arguments = [Name|_],
	command_arguments(Name, _, _),
	throw(translate_error('Wrong number of arguments.')).

translate_command(_String, Arguments, _Command) :-
	% Checks if string contains the solve command with an invalid query
	Arguments = [sv|_],
	throw(translate_error('Syntax error in query.')).

translate_command(String, _Arguments, Command) :-
	% Checks if string contains a query
	translator:translate_query(String, Query, Bindings, Degree),
	Command = sv(Query, Bindings, Degree).

translate_command(_String, _Arguments, _Command) :-
	% String contains an unknown command or bad query
	throw(translate_error('Syntax error in command or query.')).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DCG (Definite Clause Grammar) rules for parsing shell commands
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% command(-Command)
%
%     DCG rule which parses a full Bousi-Prolog shell command and
%     returns it as an executable term, using command name as functor
%     and parameters and/or options as arguments.
%
%     For example, 'qt' is translated into 'qt'; 'hp pwd' is returned
%     as 'hp(pwd)'; and 'ld -f test' is translated into 'ld(test,
%     [f])'.
%

command(Command) -->
	% Command with no arguments
	[Name],
	{
		command_arguments(Name, 0, no),
		Command =.. [Name]
	}.

command(Command) -->
	% Command with a list of options and one or more arguments
	[Name], option_lists(Options), arguments(Args),
	{
		length(Args, ArgsCount),
		command_arguments(Name, ArgsCount, yes),
		append(Args, [Options], ArgsAndOptions),
		Command =.. [Name|ArgsAndOptions]
	}.

command(Command) -->
	% Command with one or more arguments
	[Name], arguments(Args),
	{
		length(Args, NumberArgs),
		command_arguments(Name, NumberArgs, no),
		Command =.. [Name|Args]
	}.


%% option_lists(-OptionList)
%
%     DCG rule which parses zero, one or several option lists preceded
%     by a dash '-' and returns a list with all the characters found in
%     the options lists.
%

option_lists(OptionList) -->
	[Options],
	{
		% Checks that Options starts with '-' and
		% gets characters of option list
		atom_chars(Options, Chars), Chars = ['-'|OptionChars]
	},
	option_lists(MoreOptions),
	{
		% Merges option lists
		append(OptionChars, MoreOptions, OptionList)
	}.

option_lists([]) -->
	% No more option lists
	[].


%% arguments(-ArgList)
%
%     DCG rule which parses zero, one or several command arguments and
%     returns them in a list.
%

arguments(ArgList) -->
	[Arg], arguments(MoreArgs),
	{
		ArgList = [Arg|MoreArgs]
	}.

arguments([]) -->
	[].



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for loading program files
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% load_file(+Filename, +OverwriteTPL)
%
%     Loads a Bousi-Prolog program file into the Prolog database. If
%     the associated TPL file exists and is newer than the BPL file,
%     the TPL will be loaded, unless OverwriteTPL is set to 'yes'.
%

%% load_file(+BPLFilename, +TPLFilename, +FullBPLFilename, +FullTPLFilename, +OverwriteTPL)
%
%     Internal predicate called by load_file/2, which takes the full
%     and base filenames of both the program file and the destination
%     TPL file.
%
%     @see load_file/2
%

load_file(Filename, OverwriteTPL) :-
	% Backups the system flags so they can be restored if loading fails
	flags:backup_bpl_flags,
	% Gets the default BPL and TPL filenames
	get_bpl_tpl_filenames(Filename, Directory, _BaseFilename,
	                      BPLFilename, TPLFilename),
	concat_atom([Directory, '/', BPLFilename], FullBPLFilename),
	concat_atom([Directory, '/', TPLFilename], FullTPLFilename),
	% Calls the internal predicate
	load_file(BPLFilename, TPLFilename,
	          FullBPLFilename, FullTPLFilename,
	          OverwriteTPL).

load_file(BPLFilename, TPLFilename,
          FullBPLFilename, FullTPLFilename,
	      _OverwriteTPL) :-
	exists_file(FullBPLFilename),
	not(exists_file(FullTPLFilename)),
	% BPL file exists but TPL doesn't, so BPL file must be translated
	load_bpl(BPLFilename, TPLFilename,
             FullBPLFilename, FullTPLFilename).

load_file(BPLFilename, TPLFilename,
          FullBPLFilename, FullTPLFilename,
          yes) :-
	exists_file(FullBPLFilename),
	exists_file(FullTPLFilename),
	% BPL and TPL files exist, but TPL file must be overwritten
	delete_file(FullTPLFilename),
	writef('\'%w\' exists and will be overwritten.', [TPLFilename]), nl,
	load_file(BPLFilename, TPLFilename,
              FullBPLFilename, FullTPLFilename,
	          no).

load_file(BPLFilename, TPLFilename,
          FullBPLFilename, FullTPLFilename,
          no) :-
	exists_file(FullBPLFilename),
	exists_file(FullTPLFilename),
	% BPL and TPL files exist, and TPL can be loaded if it's newer than BPL
	(utilities:file_is_newer(FullTPLFilename, FullBPLFilename) ->
		% TPL file is newer and can be loaded without recompiling
		writef('\'%w\' already exists and is being loaded...', [TPLFilename]), nl,
		load_tpl(FullBPLFilename, FullTPLFilename)
	;
		% TPL file is older and must be overwritten
		delete_file(FullTPLFilename),
		writef('\'%w\' is older and will be overwritten.', [TPLFilename]), nl,
		load_bpl(BPLFilename, TPLFilename,
                 FullBPLFilename, FullTPLFilename)
	).

load_file(BPLFilename, TPLFilename,
          FullBPLFilename, FullTPLFilename,
	      _OverwriteTPL) :-
	not(exists_file(FullBPLFilename)),
	exists_file(FullTPLFilename),
	% TPL file exists but BPL doesn't, so TPL file must be loaded
	% (a warning is shown because the original file doesn't exist)
	writef('WARNING: \'%w\' does not exist.', [BPLFilename]), nl,
	writef('\'%w\' already exists and is being loaded...', [TPLFilename]), nl,
	load_tpl(FullBPLFilename, FullTPLFilename).

load_file(BPLFilename, _TPLFilename,
          FullBPLFilename, FullTPLFilename, _OverwriteTPL) :-
	not(exists_file(FullBPLFilename)),
	not(exists_file(FullTPLFilename)),
	% None of BPL and TPL files exist
	writef('ERROR: \'%w\' does not exist.', [BPLFilename]), nl.


%% load_bpl(+BPLFilename, +TPLFilename, +FullBPLFilename, +FullTPLFilename)
%
%     Parses and translates a BPL source-code file, converts it into an
%     intermediate TPL file, and then loads the latter file.
%

load_bpl(BPLFilename, TPLFilename,
         FullBPLFilename, FullTPLFilename) :-
	flags:reset_bpl_flags,
	writef('Parsing and translating \'%w\'...', [BPLFilename]), nl,
	translator:translate_program(FullBPLFilename, '', FullTPLFilename),
	writef('\'%w\' is being loaded...', [TPLFilename]), nl,
	load_tpl(FullBPLFilename, FullTPLFilename).

load_bpl(_BPLFilename, _TPLFilename,
         _FullBPLFilename, _FullTPLFilename) :-
	% This is only executed if translate_program/3 fail
	flags:restore_bpl_flags,
	write('Program not loaded.'), nl.


%% load_tpl(+FullBPLFilename, +FullTPLFilename)
%
%     Loads an intermediate TPL file into the 'evaluator' module and
%     saves the path of the new program loaded.
%

load_tpl(FullBPLFilename, FullTPLFilename) :-
	flags:reset_bpl_flags,
	evaluator:load_tpl(FullTPLFilename),
	set_files_loaded(FullBPLFilename, ''),
	flags:get_bpl_flag(program_prefix(ProgramPrefix)),
	set_program_predicates(ProgramPrefix),
	write('Program loaded!'), nl.

load_tpl(_FullBPLFilename, _FullTPLFilename) :-
	% This is only executed if evaluator:load_tpl/1 fails
	flags:restore_bpl_flags,
	write('Program not loaded.'), nl.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for loading ontology files
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% load_ontology(+Filename, +OverwriteTPL)
%
%     Loads a Bousi-Prolog ontology file into the Prolog database. If
%     no Bousi-Prolog program has been loaded before, an error is shown.
%     If the associated TPL file exists and is newer than both the
%     ontology file and the currently loaded program, the TPL will be
%     loaded, unless OverwriteTPL is set to 'yes'.
%

%% load_ontology(+BPLFilename, +ONTFilename, +TPLFilename, +FullBPLFilename, +FullONTFilename, +FullTPLFilename, +OverwriteTPL)
%
%     Internal predicate called by load_ontology/2, which takes the
%     full and base filenames of the currently loaded program, the
%     ontology file and the destintation TPL file.
%
%     @see load_ontology/2
%

load_ontology(Filename, OverwriteTPL) :-
	% Checks that a programs has been loaded before
	last_program_loaded(ProgramPath, _OntologyPath),
	ProgramPath \== '',
	% Backups the system flags so they can be restored if loading fails
	flags:backup_bpl_flags,
	% Gets the default BPL, TPL and ONT filenames
	get_bpl_tpl_filenames(ProgramPath, Directory, BaseFilename,
	                      BPLFilename, _UnusedTPLFilename),
	get_ont_filename(Filename, ONTDirectory, BaseONTFilename, ONTFilename),
	get_ontology_tpl_filename(BaseFilename, BaseONTFilename, TPLFilename),
	concat_atom([Directory, '/', BPLFilename], FullBPLFilename),
	concat_atom([ONTDirectory, '/', ONTFilename], FullONTFilename),
	concat_atom([Directory, '/', TPLFilename], FullTPLFilename),
	% Calls the internal predicate
	load_ontology(BPLFilename, ONTFilename, TPLFilename,
                  FullBPLFilename, FullONTFilename, FullTPLFilename,
	              OverwriteTPL).

load_ontology(_Filename, _OverwriteTPL) :-
	last_program_loaded('', _),
	write('ERROR: A program must be loaded before loading an ontology.'), nl.

load_ontology(BPLFilename, ONTFilename, TPLFilename,
              FullBPLFilename, FullONTFilename, FullTPLFilename,
	          _OverwriteTPL) :-
	exists_file(FullBPLFilename),
	exists_file(FullONTFilename),
	not(exists_file(FullTPLFilename)),
	% BPL and ontology files exist but TPL doesn't, so BPL + ontology must
	% be translated
	load_bpl_with_ontology(BPLFilename, ONTFilename, TPLFilename,
                           FullBPLFilename, FullONTFilename, FullTPLFilename).

load_ontology(BPLFilename, ONTFilename, TPLFilename,
              FullBPLFilename, FullONTFilename, FullTPLFilename,
	          yes) :-
	exists_file(FullBPLFilename),
	exists_file(FullONTFilename),
	exists_file(FullTPLFilename),
	% BPL, TPL and ontology files exist, but TPL file must be overwritten
	delete_file(FullTPLFilename),
	writef('\'%w\' exists and will be overwritten.', [TPLFilename]), nl,
	load_bpl_with_ontology(BPLFilename, ONTFilename, TPLFilename,
                           FullBPLFilename, FullONTFilename, FullTPLFilename).

load_ontology(BPLFilename, ONTFilename, TPLFilename,
              FullBPLFilename, FullONTFilename, FullTPLFilename,
	          no) :-
	exists_file(FullBPLFilename),
	exists_file(FullONTFilename),
	exists_file(FullTPLFilename),
	% BPL, TPL and ontology files exist, and TPL can be loaded if it's newer
	% than both BPL and ontology
	((utilities:file_is_newer(FullTPLFilename, FullBPLFilename),
	  utilities:file_is_newer(FullTPLFilename, FullONTFilename)) ->
		% TPL file is newer than ontology and BPL, so it can be loaded
		% without recompiling
		writef('\'%w\' already exists and is being loaded...', [TPLFilename]), nl,
		load_tpl_with_ontology(FullBPLFilename, FullONTFilename, FullTPLFilename)
	;
		% TPL file is older than ontology or BPL and must be overwritten
		delete_file(FullTPLFilename),
		writef('\'%w\' is older and will be overwritten.', [TPLFilename]), nl,
		load_bpl_with_ontology(BPLFilename, ONTFilename, TPLFilename,
		                       FullBPLFilename, FullONTFilename, FullTPLFilename)
	).	

load_ontology(_BPLFilename, ONTFilename, _TPLFilename,
              FullBPLFilename, FullONTFilename, _FullTPLFilename,
	          _OverwriteTPL) :-
	exists_file(FullBPLFilename),
	not(exists_file(FullONTFilename)),
	% Original BPL file exists but ontology doesn't
	writef('ERROR: \'%w\' ontology do not exist.', [ONTFilename]), nl.

load_ontology(_BPLFilename, _ONTFilename, _TPLFilename,
              FullBPLFilename, _FullONTFilename, _FullTPLFilename,
	          _OverwriteTPL) :-
	not(exists_file(FullBPLFilename)),
	% Original BPL file doesn't exist
	write('ERROR: Program\'s original BPL file is needed to load an ontology.'), nl.


%% load_bpl_with_ontology(+BPLFilename, +ONTFilename, +TPLFilename, +FullBPLFilename, +FullTPLFilename, +FullONTFilename)
%
%     Parses and translates a BPL program together with a BPL ontology,
%     converts them into a single intermediate TPL file, and then loads
%     the latter file.
%

load_bpl_with_ontology(BPLFilename, ONTFilename, TPLFilename,
                       FullBPLFilename, FullONTFilename, FullTPLFilename) :-
	flags:reset_bpl_flags,
	writef('Parsing and translating \'%w\' using ontology \'%w\'...',
	       [BPLFilename, ONTFilename]), nl,
	translator:translate_program(FullBPLFilename, FullONTFilename, FullTPLFilename),
	writef('\'%w\' is being loaded...', [TPLFilename]), nl,
	load_tpl_with_ontology(FullBPLFilename, FullONTFilename, FullTPLFilename).

load_bpl_with_ontology(_BPLFilename, _ONTFilename, _TPLFilename,
                       _FullBPLFilename, _FullONTFilename, _FullTPLFilename) :-
	% This is only executed if translate_program/3 fail
	flags:restore_bpl_flags,
	write('Ontology not loaded.'), nl.


%% load_tpl_with_ontology(+FullBPLFilename, +FullONTFilename, +FullTPLFilename)
%
%     Loads an intermediate TPL file compiled from a BPL and an
%     ontology into the 'evaluator' module and saves the path of the
%     new program and ontology loaded.
%

load_tpl_with_ontology(FullBPLFilename, FullONTFilename, FullTPLFilename) :-
	flags:reset_bpl_flags,
	evaluator:load_tpl(FullTPLFilename),
	set_files_loaded(FullBPLFilename, FullONTFilename),
	flags:get_bpl_flag(program_prefix(ProgramPrefix)),
	set_program_predicates(ProgramPrefix),
	write('Ontology loaded!'), nl.

load_tpl_with_ontology(_FullBPLFilename, _FullONTFilename, _FullTPLFilename) :-
	% This is only executed if evaluator:load_tpl/1 fails
	flags:restore_bpl_flags,
	write('Ontology not loaded.'), nl.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for handling filenames
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% get_bpl_tpl_filenames(+Filename, -Directory, -BaseFilename, -BPLFilename, -TPLFilename)
%
%     Splits a Filename into its absolute path and its base (relative)
%     filename, and then adds the '.bpl' and '.tpl' extension to the
%     base filename.
%
%     For example, given Filename = '/example/file.bpl', this predicate
%     will return Directory = '/example', BaseFilename = 'file',
%     BPLFilename = 'file.bpl' and TPLFilename = 'file.tpl'.
%

get_bpl_tpl_filenames(Filename, Directory, BaseFilename, BPLFilename, TPLFilename) :-
	% Converts filename into an absolute path and extracts
	% its relative filename and directory 
	absolute_file_name(Filename, AbsoluteFilename),
	file_base_name(AbsoluteFilename, RelativeFilename),
	file_directory_name(AbsoluteFilename, Directory),
	% Checks filename extension
	file_name_extension(FilenameNoExtension, Extension, RelativeFilename),
	downcase_atom(Extension, LowerCaseExtension),
	((LowerCaseExtension == 'bpl' ; LowerCaseExtension == 'tpl' ;
	  LowerCaseExtension == '') ->
		% Base filename from "file.bpl" or "file." is "file"
		BaseFilename = FilenameNoExtension
	;
		% Base filename from "file.new" is "file.new"
		BaseFilename = RelativeFilename
	),
	% Builds BPL and TPL filenames
	concat_atom([BaseFilename, '.', bpl], BPLFilenameAux),
	concat_atom([Directory, '/', BPLFilenameAux], FullBPLFilenameAux),
	concat_atom([Directory, '/', BaseFilename], FullBaseFilaname),
	((exists_file(FullBPLFilenameAux) ; not(exists_file(FullBaseFilaname))) ->
		BPLFilename = BPLFilenameAux
	;
		BPLFilename = BaseFilename
	),
	concat_atom([BaseFilename, '.', tpl], TPLFilename).


%% get_ont_filename(+Filename, -Directory, -BaseFilename, -ONTFilename)
%
%     Splits a Filename into its absolute path and its base (relative)
%     filename, and then adds the '.ont' extension to the base
%     filename.
%
%     For example, given Filename = '/example/ontology.ont', this
%     predicate will return Directory = '/example', BaseFilename =
%     'ontology', ONTFilename = 'ontology.ont'.
%

get_ont_filename(Filename, Directory, BaseFilename, ONTFilename) :-
	% Converts filename into an absolute path and extracts
	% relative filename and directory 
	absolute_file_name(Filename, AbsoluteFilename),
	file_base_name(AbsoluteFilename, RelativeFilename),
	file_directory_name(AbsoluteFilename, Directory),
	% Checks filename extension
	file_name_extension(FilenameNoExtension, Extension, RelativeFilename),
	downcase_atom(Extension, LowerCaseExtension),
	((LowerCaseExtension == 'ont' ; LowerCaseExtension == '') ->
		% Base filename from "file.ont" or "file." is "file"
		BaseFilename = FilenameNoExtension
	;
		% Base filename from "file.new" is "file.new"
		BaseFilename = RelativeFilename
	),
	% Builds ONT filename
	concat_atom([BaseFilename, '.', ont], ONTFilenameAux),
	concat_atom([Directory, '/', ONTFilenameAux], FullONTFilenameAux),
	concat_atom([Directory, '/', BaseFilename], FullBaseFilename),
	((exists_file(FullONTFilenameAux) ; not(exists_file(FullBaseFilename))) ->
		ONTFilename = ONTFilenameAux
	;
		ONTFilename = BaseFilename
	).


%% get_ontology_tpl_filename(+ProgramBaseFilename, +OntologyBaseFilename, -TPLFilename)
%
%     Concatenates the base filenames of a program and an ontology and
%     then adds the '.tpl' extension to the resulting filename.
%
%     For example, given ProgramBaseFilename = 'file' and
%     OntologyBaseFilename = 'ontology', this predicate will return
%     TPLFilename = 'file-ontology.tpl'.
%

get_ontology_tpl_filename(ProgramBaseFilename, OntologyBaseFilename, TPLFilename) :-
	concat_atom([ProgramBaseFilename, '-', OntologyBaseFilename, '.', tpl], TPLFilename).

	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for solving a query
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% sv(+Query, +Bindings, ?Degree)
%
%     Tries to solve the specified Bousi-Prolog Query and, if an answer
%     is found, shows the approximation degree of the result and the
%     list of variable bindings. Afterwards, user can ask for more
%     answers.
%

%% sv_aux(+Query, +Bindings, ?Degree)
%
%     Internal predicate used by sv/3.
%

sv(Query, Bindings, Degree) :-
	sv_aux(Query, Bindings, Degree),
	% This cut is used to discard all remaining choice points of the
	% query, since it's not going to be solved again; in other words,
	% this cut "frees" all memory used by the query
	!.

sv_aux(Query, [], Degree) :- 
	% Ground query (i.e., query has no free variables)
	evaluator:solve_goal(Query),
	write('Yes'), nl,
	writef('With approximation degree: %w ', [Degree]),
	sv_actions.

sv_aux(Query, Bindings, Degree) :- 
	% Non ground query (i.e., query has one or more free variables)
	Bindings \== [],
	evaluator:solve_goal(Query),
	print_answer(Bindings), nl,
	writef('With approximation degree: %w ', [Degree]),
	sv_actions.

sv_aux(_Query, _Bindings, _Degree) :-
	% Query couldn't be solved or there're no more answers
	write('No answers'), nl.


%% sv_actions
%
%     Prompts the user for an action after solving a query. User can
%     ask for more answers or finish the query.
%

sv_actions :-
	flush_output,
	get_single_char(Code),
	char_code(Char, Code),
	(member(Char, [';', 'n', 'r', ' ', '\t']) ->
		% Tries to solve the query again
		write(';'), nl,
		fail
	;
	(member(Char, ['c', 'a', '\r', '\n']) ->
		% Finishes query
		write('.'), nl,
		write('Yes'), nl
	;
	(member(Char, ['h', '?']) ->
		% Shows available commands
		nl, nl,
		write('Available actions:'), nl,
		write(' ; (n, r, space, tab): redo'), nl,
		write(' c (a, return): exit'), nl,
		write(' h (?): help'), nl, nl,
		write('Action? '), sv_actions
	;
		% Invalid action
		nl,
		write('Unknown action: '), write(Char), write(' (h for help)'), nl,
		write('Action? '), sv_actions
	))).


%% print_answer(+Bindings)
%
%     Prints the contents of a list of variable bindings that
%     represents the answer of a query. After writing the last binding,
%     a newline is not added in case user wants to ask for more
%     answers.
%

print_answer([]).

print_answer([Name = Value]) :-
	write(Name), write(' = '), write(Value).

print_answer([Name = Value|MoreBindings]) :-
	MoreBindings \== [],
	print_answer([Name = Value]), nl,
	print_answer(MoreBindings).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Shell command predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% hp
%
%     Shows some general help about Bousi-Prolog.
%

hp :-
	bplHelp:bpl_help.


%% hp(+Topic)
%
%     Shows some help about a specific topic or command.
%

hp(Topic) :-
	bplHelp:command_help(Topic).

hp(_Topic) :-
	write('ERROR: Unknown topic.'), nl.


%% qt
%
%     Stops Bousi-Prolog execution.
%

qt :-
	write('Exiting the system...'), nl, nl,
	save_history,
	halt.


%% ls
%
%     Lists the contents of current working directory, using the ls/1
%     SWI-Prolog predefined predicate.
%

ls :-
	shell:ls.


%% pwd
%
%     Prints the full path of the current working directory.
%

pwd :-
	working_directory(CurrentWorkingDir, CurrentWorkingDir),
	write('Current working directory is:'), nl,
	write(CurrentWorkingDir), nl.


%% cd(+Directory)
%
%     Changes current working directory to Directory, which can be an
%     absolute or a relative path.
%

cd(Directory) :- 	
	exists_directory(Directory), !,
	working_directory(_OldDirectory, Directory),
	working_directory(NewWorkingDir, NewWorkingDir),
	write('New working directory is:'), nl,
	write(NewWorkingDir), nl.

cd(Directory) :-
	writef('ERROR: Directory \'%w\' does not exist.', [Directory]), nl.


%% sh
%
%     Starts an interactive command-line shell. On Unix/Linux, default
%     shell is /bin/sh, but can be overriden with $SHELL environment
%     variable; on Windows, default shell is cmd.exe, but can be
%     overriden with %COMSPEC% environment variable.
%

sh :-
	current_prolog_flag(windows, true), !,
	write('Starting a Windows shell...'), nl,
	write('Type \'exit\' to return to Bousi~Prolog.'), nl,
	(getenv('COMSPEC', Shell), ! ; Shell = 'cmd.exe'),
	% shell/2 predicate works well on Windows Vista and 7, but hangs on
	% XP, that's because we use win_shell/2 instead, the only difference
	% being that win_shell/2 doesn't wait until task ends
	win_shell(open, Shell).

sh :-
	write('Starting a Unix shell...'), nl,
	write('Type \'exit\' to return to Bousi~Prolog.'), nl, nl,
	(getenv('SHELL', Shell), ! ; Shell = '/bin/sh'),
	shell(Shell, _).


%% lc
%
%     Shows current lambda-cut value, i.e., the lower bound allowed for
%     the resulting approximation degree of a weak unification.
%

lc :-
	flags:get_bpl_flag(lambda_cut(LambdaValue)),
	writef('Current lambda-cut value is: %w', [LambdaValue]), nl.


%% lc(+Lambda)
%
%     Changes current lambda-cut value, i.e., the lower bound allowed
%     for the resulting approximation degree of a weak unification.
%

lc(Lambda) :-
	catch((
		% Converts Lambda atom into a number
		atom_chars(Lambda, LambdaChars),
		number_chars(LambdaValue, LambdaChars),
		LambdaValue >= 0, LambdaValue =< 1, !,
		% Changes lambda-cut value
		flags:remove_bpl_flag(lambda_cut(_OldLambda)),
		flags:add_bpl_flag(lambda_cut(LambdaValue)),
		writef('New lambda-cut value is: %w', [LambdaValue]), nl
	% (catcher)
	), error(syntax_error(illegal_number), _), (
		% Lambda-cut is not a number
		fail
	)).

lc(Lambda) :-
	writef('ERROR: \'%w\' is a wrong lower bound for the approximation \c
	        degree.', [Lambda]), nl,
	write('Lambda-cut value must be a number in range [0.0, 1.0].'), nl.


%% ld
%
%     Shows the name of the program and the ontology that are currently
%     loaded into memory.
%

ld :-
	last_program_loaded('', _), !,
	write('No program loaded.'), nl.

ld :-
	last_program_loaded(Program, ''), !,
	write('Current loaded program is:'), nl,
	write(Program), nl.

ld :-
	last_program_loaded(Program, Ontology),
	write('Current loaded program is:'), nl,
	write(Program), nl,
	write('Current loaded ontology is:'), nl,
	write(Ontology), nl.


%% ld(+Filename)
%
%     Loads a Bousi-Prolog source code file into memory. If TPL file
%     exists and is newer than BPL file, TPL will be loaded.
%

ld(Filename) :-
	load_file(Filename, no).


%% ld(+Filename, +Options)
%
%     Loads a Bousi-Prolog source code file into memory as a program,
%     or as an ontology if 'o' option is specified. If TPL file exists
%     and is newer than source code file, TPL will be loaded, unless
%     the 'f' option is specified.
%

ld(Filename, Options) :-
	is_list(Options),
	subset(Options, [f, o]),
	(member(f, Options) ->
		OverwriteTPL = yes
	;
		OverwriteTPL = no
	),
	(member(o, Options) ->
		load_ontology(Filename, OverwriteTPL)
	;
		load_file(Filename, OverwriteTPL)
	).

ld(_Filename, Options) :-
	member(Option, Options),
	not(member(Option, [f, o])),
	writef('ERROR: Unknown option: \'%w\'.', [Option]), nl.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for reading and writing command history
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% load_history
%
%     Reads the command history used by editline library from the
%     '.bpl_history' file located in the user home directory.
%

load_history :-
	history_filename(File),
	utilities:home_directory(HomeDirectory),
	concat_atom([HomeDirectory, '/', File], HistoryFile),
	(exists_file(HistoryFile) ->
		true
	;
		% Creates a new, empty history file
		telling(CurrentOutput),
		tell(HistoryFile), told,
		tell(CurrentOutput)
	),
	foreign:ext_load_shell_history(HistoryFile).


%% save_history
%
%     Writes the command history used by editline library in the
%     '.bpl_history' file located in the user home directory.
%

save_history :-
	history_filename(File),
	max_history_commands(MaxCommands),
	utilities:home_directory(HomeDirectory),
	concat_atom([HomeDirectory, '/', File], HistoryFile),
	foreign:ext_save_shell_history(HistoryFile, MaxCommands).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for managing lists of predicate names
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% set_system_predicates
%
%     Retrieves the list of the names of the predefined predicates
%     in SWI-Prolog and sets them to be used by the shell's
%     autocomplete feature.
%

set_system_predicates :-
	utilities:get_predicates(PredicateNames),
	foreign:ext_set_system_predicate_list(PredicateNames).


%% set_program_predicates(+Prefix)
%
%     Retrieves the list of the names of the predicates defined in the
%     currently loaded program and sets them to be used by the shell's
%     autocomplete feature. Prefix represents the prefix associated to
%     the loaded program.
%

set_program_predicates(Prefix) :-
	utilities:get_predicates_modules([evaluator], PredicateNames),
	atom_concat(Prefix, '_', FullPrefix),
	utilities:remove_prefixes(PredicateNames, RealNames, FullPrefix),
	foreign:ext_set_program_predicate_list(RealNames).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Miscellaneous predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% set_files_loaded(+File, +Ontology)
%
%     Sets File and Ontology as the new paths of the Bousi-Prolog
%     program and ontology that are currently loaded into memory.
%     Besides, updates the program_prefix flag from File.
%

set_files_loaded(File, Ontology) :-
	last_program_loaded(OldFile, OldOntology),
	retract(last_program_loaded(OldFile, OldOntology)),
	assert(last_program_loaded(File, Ontology)),
	utilities:simplify_filename(File, SimplifiedFilename),
	flags:remove_bpl_flag(program_prefix(_OldProgramPrefix)),
	flags:add_bpl_flag(program_prefix(SimplifiedFilename)).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Constant predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% bpl_prompt(?Prompt)
%
%     Unifies Prompt with the prompt that will be shown to the user.
%

bpl_prompt('BPL> ').


%% history_filename(?Filename)
%
%     Unifies Filename with the name of the file where the Bousi-Prolog
%     command history will be saved.
%

history_filename('.bpl_history').


%% max_history_commands(?Max)
%
%     Unifies Max with the maximum number of commands that may be saved
%     in the Bousi-Prolog command history.
%

max_history_commands(100).


%% command_arguments(?Command, ?NArgs, ?Options)
%
%     For each Bousi-Prolog shell command (except sv), unifies Command
%     with the command name, NArgs with the number of required
%     arguments, and Options with 'yes' if the command can take a list
%     of options or 'no' otherwise.
%

command_arguments(hp, 0, no).
command_arguments(lc, 0, no).
command_arguments(ld, 0, no).
command_arguments(ls, 0, no).
command_arguments(pwd, 0, no).
command_arguments(qt, 0, no).
command_arguments(sh, 0, no).
command_arguments(cd, 1, no).
command_arguments(hp, 1, no).
command_arguments(lc, 1, no).
command_arguments(ld, 1, no).
command_arguments(ld, 1, yes).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% last_program_loaded(?Program, ?Ontology)                  is dynamic
%
%     Dynamic predicate which stores the full paths of the Bousi-Prolog
%     Program and Ontology that are currently loaded into memory.
%     Program and Ontology can be '' (an empty string) if no program
%     has been loaded yet, whereas Ontology can also be '' if just a
%     program is currently loaded.
%

:- dynamic last_program_loaded/2.

last_program_loaded('', '').

