%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% System flags handler

:- module(flags, [
		get_bpl_flag/1,         % ?Flag
		remove_bpl_flag/1,      % +Flag
		add_bpl_flag/1,         % +Flag
		reset_bpl_flags/0,      %
		backup_bpl_flags/0,     %
		restore_bpl_flags/0     %
   ]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for handling system flags
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% get_bpl_flag(?Flag)
%
%     Queries current system flags.
%

get_bpl_flag(Flag) :-
	bpl_flags(Flag).


% add_bpl_flag(+Flag)
%
%     Adds a new Flag to the current system flags. Does nothing if
%     Flag already belongs to the system flags.
%

add_bpl_flag(Flag) :-
	bpl_flags(Flag),
	% Flag already exists
	!.

add_bpl_flag(Flag) :-
	% Flag doesn't exist and must be added
	assert(bpl_flags(Flag)).


%% remove_bpl_flag(+Flag)
%
%     Removes a Flag from the current system flags. Does nothing if
%     Flag doesn't belong to the system flags.
%

remove_bpl_flag(Flag) :-
	bpl_flags(Flag),
	% Flag exists and must be removed
	!,
	retract(bpl_flags(Flag)).

remove_bpl_flag(_Flag).
	% Flag doesn't exist


%% backup_bpl_flags
%
%     Saves a copy of the current system flags that can be later
%     restored with restore_bpl_flags/0.
%
%     @see restore_bpl_flags/0
%

backup_bpl_flags :-
	% Removes previously saved flags
	retractall(saved_bpl_flags(_OldSavedFlags)),
	% Copies all "bpl_flags(X)" to "saved_bpl_flags(X)"
	findall(saved_bpl_flags(Flag), bpl_flags(Flag), FlagsToSave),
	maplist(assert, FlagsToSave).


%% restore_bpl_flags
%
%     Restores the system flags that were previously saved with
%     backup_bpl_flags/0.
%
%     @see backup_bpl_flags/0
%

restore_bpl_flags :-
	% Removes current system flags
	retractall(bpl_flags(_OldFlags)),
	% Copies all "saved_bpl_flags(X)" to "bpl_flags(X)"
	findall(bpl_flags(Flag), saved_bpl_flags(Flag), FlagsToRestore),
	maplist(assert, FlagsToRestore).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for reseting system flags
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% reset_bpl_flags
%
%     Sets the default values for all BPL flags.
%

reset_bpl_flags :-
	reset_program_prefix,
	reset_lambda_cut,
	reset_relation_properties([sim, lEqThan, gEqThan, frel1, frel2, frel3]),
	reset_fuzzy_subsets.


%% reset_program_prefix
%
%     Sets a default program name as the prefix of the currently loaded
%     program.
%

reset_program_prefix :-
	default_program_prefix(Prefix),
	remove_bpl_flag(program_prefix(_OldPrefix)),
	add_bpl_flag(program_prefix(Prefix)).


%% reset_lambda_cut(+Lambda)
%
%     Sets the default lambda-cut value in BPL flags.
%

reset_lambda_cut :-
	default_lambda(Lambda),
	remove_bpl_flag(lambda_cut(_OldLambda)),
	add_bpl_flag(lambda_cut(Lambda)).


%% reset_relation_properties(+Relation)
%
%     Sets the default closure properties and t-norm in BPL flags for
%     the specified Relation. If Relation is a list, this predicate
%     will be called once for each item.
%

reset_relation_properties([]) :-
	!.

reset_relation_properties([RelName|OtherRelNames]) :-
	% Calls this predicate recursively for each item
	reset_relation_properties(RelName),
	reset_relation_properties(OtherRelNames).

reset_relation_properties(RelName) :-
	% Gets default closure properties, converts them into a number
	% and stores it in BPL flags along with t-norm identifier
	atom(RelName),
	default_closure(RelName, ClosureProperties),
	remove_bpl_flag(relation_properties(RelName, _OldClosureProperties)),
	add_bpl_flag(relation_properties(RelName, ClosureProperties)).


%% reset_fuzzy_sets
%
%     Removes all domains and fuzzy subsets from BPL flags.
%

reset_fuzzy_subsets :-
	retractall(bpl_flags(fuzzy_domain(_DomainName1, _Definition))),
	retractall(bpl_flags(fuzzy_subsets(_DomainName2, _Subsets))).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Default values for system flags
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% default_program_prefix(?Prefix)
%
%     Unifies Prefix with the default prefix that will be appended to
%     user-defined predicates when no program is loaded.
%

default_program_prefix('none').


%% default_lambda(?Value)
%
%     Unifies Value with the default lambda-cut value, i.e., the lower
%     bound allowed for the approximation degree of weak unifications.
%     By default, this number is 0, so none of the computations are cut
%     regardless of their approximation degree.
%

default_lambda(0).


%% default_closure(?Relation, ?Properties)
%
%     Returns a list with the names of the default closure Properties
%     of a certain Relation. By default, a proximity relation is used
%     in weak unifications; partial orders are applied in 'more/less
%     general than' fuzzy relations; and custom fuzzy binary relations
%     are defined as similarity relations.
%

default_closure(SimRelation, [symmetric, reflexive]) :-
	member(SimRelation, [sim]).

default_closure(GeneralRelation, [reflexive, transitive(TNorm)])  :-
	member(GeneralRelation, [gEqThan, lEqThan]),
	default_tnorm(GeneralRelation, TNorm).

default_closure(CustomRelation, [symmetric, reflexive, transitive(TNorm)])  :-
	member(CustomRelation, [frel1, frel2, frel3]),
	default_tnorm(CustomRelation, TNorm).


%% default_tnorm(?Relation, ?TNorm)
%
%     Returns the name of the default TNorm to be used to compute the
%     transitive closure of a certain Relation. 'Minimum' is the
%     default t-norm in all the relations except in the one used in the
%     weak unification process, where no transitivity is applied.
%

default_tnorm(sim, no).
default_tnorm(gEqThan, min).
default_tnorm(lEqThan, min).
default_tnorm(frel1, min).
default_tnorm(frel2, min).
default_tnorm(frel3, min).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% bpl_flags(?Flag)                                          is dynamic
%
%     Dynamic predicate which stores some flags that affect the
%     behavior of the Bousi-Prolog system. Currently, the following
%     flags are supported:
%
%      * lambda_cut(Lambda): defines the lambda-cut value, i.e., the
%        lower bound allowed for the approximation degree of weak
%        unifications.
%      * relation_properties(Name, Properties): contains the name of
%        the closure properties that should be applied when computing
%        the closure of each fuzzy relation. Valid properties are
%        "symmetric", "reflexive" and "transitive(TNorm)", where TNorm
%        can be "yes", "no", "min", "product" or "luka".
%      * fuzzy_domain(DomainName, Definition): defines the domains to
%        which belong the fuzzy subsets. Each domain definition must
%        be a list consisting of three items: [Min, Max, MeasureUnit].
%      * fuzzy_subsets(DomainName, Subsets): declares the lists of
%        fuzzy subsets that belong to each domain. Subset definitions
%        must follow the syntax of ext_translate_fuzzysets/5.
%      * program_prefix(Prefix): prefix that represents the program
%        currently loaded into memory. This prefix must be appended to
%        every call to a user-defined predicate.
%

:- dynamic bpl_flags/1.


%% saved_bpl_flags(?Flag)                                    is dynamic
%
%     Dynamic predicate used by backup_bpl_flags/0 and
%     restore_bpl_flags/0 to store a backup copy of the system flags.
%

:- dynamic saved_bpl_flags/1.

